-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

-- character fields
local FIELD_CURRENTSET = "FIELD_CURRENTSET";
local FIELD_LASTSET = "FIELD_LASTSET";
local FIELD_SETS = "FIELD_SETS";
local FIELD_EVENTS = "FIELD_EVENTS";

-- set fields
local FIELD_SHOWCLOAK = "FIELD_SHOWCLOAK";
local FIELD_SHOWHELMET = "FIELD_SHOWHELMET";
local FIELD_SLOTS = "FIELD_SLOTS";
local FIELD_SLOTSTATES = "FIELD_SLOTSTATES";
local FIELD_ACTIONSLOTS = "FIELD_ACTIONSLOTS";

-- misc fields
local FIELD_TYPE = "FIELD_TYPE";
local FIELD_ID = "FIELD_ID";
local FIELD_SUBTYPE = "FIELD_SUBTYPE";

function c:InitDatabase()
	GQ_DATA[c:GetRealmName()] = GQ_DATA[c:GetRealmName()] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()] = GQ_DATA[c:GetRealmName()][c:GetCharName()] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS] or {};
end

function c:DataBase_Update14()
	for realmName in pairs(GQ_DATA) do
		for charName in pairs(GQ_DATA[realmName]) do
			for setName in pairs(GQ_DATA[realmName][charName][FIELD_SETS]) do
				GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES] = GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES] or {};
				for slotId in pairs(c:GetSlotInfo()) do
					if GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] == nil then
						GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] = true;
					end
				end
			end
		end
	end
end

function c:LoadSetNames()
	local setNames = {};
	for setName in pairs(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS]) do
		table.insert(setNames, setName);
	end

	table.sort(setNames);
	return setNames;
end

function c:LoadCurrentSetName()
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET];
end

function c:LoadPreviousSetName()
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_LASTSET];
end

function c:SaveCurrentSetName(setName)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_LASTSET] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET];
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET] = setName;
end

function c:LoadSet(setName)
	setName = setName or c:LoadCurrentSetName();
	if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS];
	end
end

function c:SaveSet(setName, notify)
	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS] or {};

		for slotId, slotName in pairs(c:GetSlotInfo()) do
			c:SaveSlot(slotId, setName, notify);
			c:SaveSlotState(slotId, true, notify, setName);
		end

		CloseDropDownMenus();
		c:SaveCloakAndHelmet(setName, notify);
		c:SaveActionBars(setName, notify);
		c:SaveCurrentSetName(setName);
		c:RefreshSetList(setName);
		c:GetBindingEntries(true);
		c:SetPaperDollLabelText();
	end
end

function c:SaveDefaultSet()
	c:SaveSet(c:GetText("default"), false);
end

function c:DeleteSet(setName)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = nil;

	if c:LoadCurrentSetName() == setName then
		c:SaveCurrentSetName(nil);
	end
	for eventName, bindingName in pairs(c:LoadEventBindings()) do
		if setName == bindingName then
			c:SaveEventBinding(eventName, nil);
		end
	end

	CloseDropDownMenus();
	c:RefreshSetList();
	c:GetBindingEntries(true);
	c:RefreshEventEntries();
	c:SetPaperDollLabelText();
	c:HideSlotStateBoxes();
end

function c:SaveSlot(slotId, setName, notify, itemId)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	if setName then
		itemId = itemId or c:GetItemId(slotId);
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS][slotId] ~= itemId then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS][slotId] = itemId;
			if notify  then
				c:NotifyInChat(c:GetText("<%s> Saved slot %s (%s): %s", setName, slotId, c:GetDisplaySlotName(slotId), c:GetItemLink(itemId)));
			end
		end
	end
end

function c:LoadSlotState(slotId, setName)
	setName = setName or c:LoadCurrentSetName();
	if setName then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId];
	end
end

function c:SaveSlotState(slotId, active, notify, setName)
	setName = setName or c:LoadCurrentSetName();
	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] or {};
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] ~= active then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] = active;
			if notify then
				c:NotifyInChat(c:GetText("<%s> Saved slot state for %s (%s): %s", setName, slotId, c:GetDisplaySlotName(slotId), c:BoolToText(active, c:GetText("active"), c:GetText("not active"))));
			end
		end
	end
end

function c:SaveCloakAndHelmet(setName, notify)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};

		if ShowingHelm() ~= GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] = ShowingHelm();
			if notify then
				c:NotifyInChat(string.format("<%s> %s %s", setName, c:GetText("Saved helmet visibility:"), c:BoolToText(ShowingHelm(), c:GetText("visible"), c:GetText("not visible"))));
			end
		end

		if ShowingCloak() ~= GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] = ShowingCloak();
			if notify then
				c:NotifyInChat(string.format("<%s> %s %s", setName, c:GetText("Saved cloak visibility:"), c:BoolToText(ShowingCloak(), c:GetText("visible"), c:GetText("not visible"))));
			end
		end
	end
end

function c:LoadCloakAndHelmet(setName)
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] ~= nil then
			ShowHelm(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET]);
		end
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] ~= nil then
			ShowCloak(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK]);
		end
	end
end

function c:SaveActionBars(setName, notify)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] or {};

	for slotId = 1, 120 do
		c:SaveActionSlot(slotId, setName, notify);
	end
end

function c:SaveActionSlot(slotId, setName, notify)
	setName = setName or c:LoadCurrentSetName();
	if setName then
		if notify == nil then
			notify = true;
		end

		local entry = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId];
		if not c:IsSameAction(slotId, entry) then
			if HasAction(slotId) then
				local actionType, id, subType = GetActionInfo(slotId);

				local newEntry = {};
				newEntry[FIELD_TYPE] = actionType;
				newEntry[FIELD_ID] = id;
				newEntry[FIELD_SUBTYPE] = subType;

				GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] = newEntry;
			else
				GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] = c.VALUE_NONE;
			end

			if notify then
				local actionItemName = c:GetText("Empty");

				if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] ~= c.VALUE_NONE then
					local aType, id = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][FIELD_TYPE], GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][FIELD_ID];

					if aType == "spell" then
						local name, rank = c:GetSpellName(id), c:GetSpellRank(id);
						if rank and rank ~= "" then
							actionItemName = c:GetText("%s (%s)", name, rank);
						else
							actionItemName = name;
						end
					elseif aType == "item" then
						actionItemName = c:GetRarityColor(c:GetItemRarity(id))["hex"] .. c:GetItemName(id);
					elseif aType == "macro" then
						actionItemName = c:GetMacroName(id);
					end
				end
				c:NotifyInChat(c:GetText("<%s> Saved action slot %s: %s", setName, slotId, actionItemName));
			end
		end
	end
end

function c:LoadActionBars(setName, retry)
	for slotId = 1, 120 do
		local entry = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId];

		if entry and not c:IsSameAction(slotId, entry) then
			if InCombatLockdown() then
				-- breaks loop if combat starts
				if not retry then
					c:Println(c:GetText("Switching interrupted by combat. It will be re-attempted once combat is over."));
				end
				C_Timer.After(1, function() c:SwitchToSet(setName, true); end);
				return;
			end

			ClearCursor();
			PickupAction(slotId);
			ClearCursor();

			if entry ~= c.VALUE_NONE then
				if entry[FIELD_TYPE] == "spell" then
					PickupSpell(entry[FIELD_ID]);
				elseif entry[FIELD_TYPE] == "item" then
					PickupItem(entry[FIELD_ID]);
				elseif entry[FIELD_TYPE] == "macro" then
					PickupMacro(entry[FIELD_ID]);
				end

				PlaceAction(slotId);
				ClearCursor();
			end

			c:HighlightActionSlot(slotId);
		end
	end

	return true;
end

function c:IsSameAction(slotId, entry)
	if HasAction(slotId) and entry and entry ~= c.VALUE_NONE then
		local actionType, id, subType = GetActionInfo(slotId);
		if actionType == entry[FIELD_TYPE] and id == entry[FIELD_ID] and subType == entry[FIELD_SUBTYPE] then
			return true;
		end
	elseif not HasAction(slotId) and (not entry or entry == c.VALUE_NONE) then
		return true;
	end
end

function c:SaveEventBinding(eventName, setName)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS][eventName] = setName;
	c:NotifyInChat(string.format("%s %s -> %s", c:GetText("Saved event binding:"), c:GetEvents()[eventName], c:GetBindingText(setName, "TEXT_CHAT") or c:GetText("None")));
end

function c:LoadEventBinding(eventName)
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS][eventName];
end

function c:LoadEventBindings()
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS];
end