-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

GQ_OPTIONS = GQ_OPTIONS or {};
GQ_DATA = GQ_DATA or {};

local GQ_VERSION = 15;

local realmName, charName;
local switchQueue, switching = {};
local lastMountState, lastStealthState;

-- init
function c:Init()
	realmName = GetRealmName();
	charName = UnitName("player");

	c.VALUE_NONE = "VALUE_NONE";
	c.KEYWORD_NONE = "$NONE";
	c.KEYWORD_PREVIOUS = "$PREVIOUS";

	c:InitOptions();
	c:InitDatabase();
	c:InitUI();

	-- updates
	if not GQ_OPTIONS[c.OPT_VERSION] or tonumber(GQ_OPTIONS[c.OPT_VERSION]) < GQ_VERSION then
		if not GQ_OPTIONS[c.OPT_VERSION] or tonumber(GQ_OPTIONS[c.OPT_VERSION]) < 14 then
			-- add slotStates
			c:DataBase_Update14();
		end

		GQ_OPTIONS[c.OPT_VERSION] = GQ_VERSION; -- finish update
		c:Println(c:GetText("Update to version %s complete.", GQ_VERSION));
	end

	c.initFinished = true;
	c:Println(c:GetText("AddOn loaded: Version: %s. Type /gq for options.", GQ_VERSION));
	c:Println(c:GetText("Visit https://www.curseforge.com/wow/addons/gearquipper for support, feedback and updates."));
end

function c:GetVersion()
	return GQ_VERSION;
end

function c:GetCharName()
	return charName;
end

function c:GetRealmName()
	return realmName;
end

function c:NotifyInChat(msg)
	if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] then
		c:Println(msg);
	end
end

local function HighlightButton(button)
	if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] and button then
		ActionButton_ShowOverlayGlow(button);
		C_Timer.After(1, function () ActionButton_HideOverlayGlow(button); end);
	end
end

function c:HighlightInventorySlot(slotId)
	if c:GetSlotInfo()[slotId] then
		HighlightButton(_G["Character" .. c:GetSlotInfo()[slotId]]);
	end
end

function c:HighlightActionSlot(slotId)
	HighlightButton(c:GetActionButton(slotId));
end

function c:IsSwitching()
	return switching;
end

-- command handling
SLASH_GEARQUIPPER1 = "/gq";
local CMD_RESET = "reset";
local CMD_SWITCH = "switch";

SlashCmdList["GEARQUIPPER"] = function(msg)
	if msg then
		if c.initFinished then
			msg = c:Trim(msg);

			if msg == CMD_RESET then
				c:ShowResetDialog();
				return;
			elseif c:StartsWith(msg, CMD_SWITCH, true) then
				msg = c:Trim(msg:gsub(CMD_SWITCH, ""));
				local existingSets, setNames = c:LoadSetNames(), {};
				if msg ~= "" then
					for v in string.gmatch(msg, "[^ ]+") do
						if string.upper(v) == c.KEYWORD_PREVIOUS then
							table.insert(setNames, c.KEYWORD_PREVIOUS);
						else
							local existingSetName = c:TableContains(existingSets, v, true);
							if existingSetName then
								table.insert(setNames, existingSetName);
							else
								c:Println(c:GetText("Macro warning: There is no set named \"%s\".", v));
							end
						end
					end
				else
					setNames = c:Deepcopy(existingSets);
				end

				local noSets = table.getn(setNames);
				if noSets == 1 then
					c:SwitchToSet(setNames[1]);
				elseif noSets > 1 then
					local currentSet = c:LoadCurrentSetName();
					for k, v in ipairs(setNames) do
						if v == currentSet then
							local index = k + 1;
							if index > noSets then
								index = 1;
							end
							c:SwitchToSet(setNames[index]);
							return;
						end
					end
					c:SwitchToSet(setNames[1]);
				end
				return;
			end

			c:Println(c:GetText("Unknown command. Possible parameters are: /gq ..."));
			c:Println(c:GetText("actionbars -> Toggles whether action bars are switched with your sets."));
			c:Println(c:GetText("reset -> Resets the GearQuipper addon (e.g. in case of errors)."));
			c:Println(c:GetText("switch [setname1] [setname2] ... -> Switches between specified sets. For macro use."));
		else
			c:Println(c:GetText("The AddOn has not been initialized due to an error. Please relog and try again."));
		end
	end
end

-- event handling
c.eventFrame = c.eventFrame or CreateFrame("Frame");
c.eventFrame:RegisterEvent("ADDON_LOADED");
c.eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
c.eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");

c.eventFrame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
c.eventFrame:RegisterEvent("ACTIONBAR_SLOT_CHANGED");

c.eventFrame:RegisterEvent("ITEM_LOCKED");
c.eventFrame:RegisterEvent("ITEM_UNLOCKED");

--c.eventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
c.eventFrame:RegisterEvent("UNIT_AURA");

c.eventFrame:HookScript("OnEvent", function(self, event, arg1, arg2, arg3, ...)
	if event then
		if event == "ADDON_LOADED" and arg1 == "GearQuipper" and not c.initFinished then
			c:Init();
		elseif event == "PLAYER_ENTERING_WORLD" then
			-- create default gear if not existing
			if table.getn(c:LoadSetNames()) == 0 then
				c:SaveDefaultSet();
			end

			if lastMountState == nil then
				lastMountState = IsMounted();
			end
			if lastStealthState == nil then
				lastStealthState = IsStealthed();
			end
		elseif event == "PLAYER_EQUIPMENT_CHANGED" and c:SaveConditionsMet() then
			if c:GetSlotInfo()[arg1] and not c:IsSwitching() then
				if c:LoadSlotState(arg1) then
					local currentSet = c:LoadSet();
					if currentSet then
						if not c:IsSetItem(arg1, currentSet[arg1]) then
							c:SaveSlot(arg1);
							c:SetSlotInfo(arg1, "green");
						end
					end
				else
					c:SetSlotInfo(arg1, "red");
				end
			end
		elseif event == "ACTIONBAR_SLOT_CHANGED" then
			if not c:IsSwitching() and GQ_OPTIONS[c.OPT_ACTIONBARS] then
				c:SaveActionSlot(arg1);
			end
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			if InCombatLockdown() then
				-- combat started
				c:DisablePaperDollButton();
			end
		elseif event == "ITEM_LOCKED" and not arg2 then
			if not c:IsSwitching() then
				local slotId = tonumber(arg1);
				if slotId >= 0 and slotId <= 19 then
					c:OpenQuickBar(slotId);
				end
			end
		elseif event == "ITEM_UNLOCKED" then
			c:CloseQuickBars();
		--elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
		elseif event == "UNIT_AURA" then
			C_Timer.After(1, function()
				if IsMounted() and not lastMountState then
					c:HandleEvent(c.EVENT_MOUNT);
				elseif lastMountState and not IsMounted() then
					c:HandleEvent(c.EVENT_DISMOUNT);
				elseif IsStealthed() and not lastStealthState then
					c:HandleEvent(c.EVENT_STEALTH);
				elseif lastStealthState and not IsStealthed() then
					c:HandleEvent(c.EVENT_UNSTEALTH);
				end
			end);
		end
	end
end);

function c:HandleEvent(eventName)
	if eventName == c.EVENT_MOUNT then
		lastMountState = true;
	elseif eventName == c.EVENT_DISMOUNT then
		lastMountState = false;
	elseif eventName == c.EVENT_STEALTH then
		lastStealthState = true;
	elseif eventName == c.EVENT_UNSTEALTH then
		lastStealthState = false;
	end

	local setBinding = c:LoadEventBinding(eventName);
	if setBinding then
		c:SwitchToSet(setBinding);
	end
end

function c:FinishSwitch(setName, retry)
	C_Timer.After(1, function()
		if retry and c:IsSetComplete(setName) then
			c:Println(c:GetText("Switch to set \"%s\" complete.", setName));
		end
		switching = nil;

		if table.getn(switchQueue) > 0 then
			local nextSet = table.remove(switchQueue, 1);
			c:SwitchToSet(nextSet);
		else
			c:LoadCloakAndHelmet(setName);			
			if PaperDollFrame:IsVisible() then
				c:SetPaperDollLabelText();
				c:SetSlotInfo();
				c:ShowSlotStateBoxes();
				c:RefreshSetList();
				c:UnlockUI();
			end
		end
	end);
end

function c:SwitchToSet(setName, retry)
	if (not c:IsSwitching() or retry) and setName then
		if setName == c.KEYWORD_PREVIOUS then
			setName = c:LoadPreviousSetName();
		end

		local set = c:LoadSet(c:TableContains(c:LoadSetNames(), setName, true)); -- case insensitive
		if set then
			local freeSpace, neededSpace, bagSpaceCache = c:CheckNeccessaryBagSpace(setName);
			if neededSpace > freeSpace then
				c:Println(c:GetText("Not enough bag space. Current space: %s, needed: %s", freeSpace, neededSpace));
				c:RefreshSetList();
			else
				switching = setName;
				c:LockUI(setName);

				for _, slotId in ipairs(c:GetSlotSwitchOrder()) do
					if not c:IsSetItem(slotId, set[slotId]) and (not GQ_OPTIONS[c.OPT_PARTIALSETS] or c:LoadSlotState(slotId, setName)) then
						if InCombatLockdown() then --and not (slotId == INVSLOT_MAINHAND or slotId == INVSLOT_OFFHAND or slotId == INVSLOT_RANGED) then
							-- breaks loop if combat starts
							if not retry then
								c:Println(c:GetText("Switching interrupted by combat. It will be re-attempted once combat is over."));
								c:RefreshSetList();
							end
							C_Timer.After(1, function() c:SwitchToSet(setName, true); end);
							return;
						end

						if set[slotId] == c.VALUE_NONE then
							-- unequip item -> bagspace - 1
							ClearCursor();
							PickupInventoryItem(slotId);
							if not c:PutInBag(bagSpaceCache) then
								c:Println(c:GetText("Not enough bag space. Switching aborted."));
								c:RefreshSetList();
								switching = nil;
								return;
							end
							ClearCursor();
						--elseif InCombatLockdown() and (slotId == INVSLOT_MAINHAND or slotId == INVSLOT_OFFHAND or slotId == INVSLOT_RANGED) then
							-- experimental
						--	c:CallSecureAction(setName);
						else
							if c:GetItemId(slotId) ~= c.VALUE_NONE then
								-- equip item on empty slot -> bagspace + 1
								local bagId = c:FindItemInBags(slotId);
								if bagId then
									bagSpaceCache[bagId] = bagSpaceCache[bagId] + 1;
								end
							end
							EquipItemByName(set[slotId], slotId);
						end

						if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] then
							c:HighlightInventorySlot(slotId);
						end
					end
				end

				if GQ_OPTIONS[c.OPT_ACTIONBARS] then
					c:LoadActionBars(setName, retry);
				end

				c:SaveCurrentSetName(setName);
				c:FinishSwitch(setName, retry);
			end
		end
	elseif setName and not retry then
		table.insert(switchQueue, setName);
	end
end

-- todo: remember last bag slot
function c:PutInBag(bagSpaceCache)
	for i = 0, NUM_BAG_SLOTS do
		if bagSpaceCache[i] > 0 then
			if i == 0 then
				PutItemInBackpack();
			else
				PutItemInBag(ContainerIDToInventoryID(i));
			end
			bagSpaceCache[i] = bagSpaceCache[i] - 1;
			return true;
		end
	end
end