-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

function c:IsValidSetName(candidate)
	return candidate and candidate ~= "" and (not c:StringContains(candidate, "$")) and (not c:TableContains(c:LoadSetNames(), candidate, true));
end

function c:GetItemId(slotId)
	if slotId then
		local itemId = GetInventoryItemID("player", slotId);
		if itemId and itemId ~= 0 then
			return itemId;
		end
	end
	return c.VALUE_NONE;
end

local ITEMINFO_NAME = "ITEMINFO_NAME";
local ITEMINFO_LINK = "ITEMINFO_LINK";
local ITEMINFO_RARITY = "ITEMINFO_RARITY";
local ITEMINFO_LEVEL = "ITEMINFO_LEVEL";
local ITEMINFO_MINLEVEL = "ITEMINFO_MINLEVEL";
local ITEMINFO_TYPE = "ITEMINFO_TYPE";
local ITEMINFO_SUBTYPE = "ITEMINFO_SUBTYPE";
local ITEMINFO_STACKCOUNT = "ITEMINFO_STACKCOUNT";
local ITEMINFO_EQUIPLOC = "ITEMINFO_EQUIPLOC";
local function GetItemInfos(itemId, info)
	if itemId and itemId ~= c.VALUE_NONE then
		local itemInfo = {};
		itemInfo[ITEMINFO_NAME], itemInfo[ITEMINFO_LINK], itemInfo[ITEMINFO_RARITY], itemInfo[ITEMINFO_LEVEL],
		itemInfo[ITEMINFO_MINLEVEL], itemInfo[ITEMINFO_TYPE], itemInfo[ITEMINFO_SUBTYPE], itemInfo[ITEMINFO_STACKCOUNT],
		itemInfo[ITEMINFO_EQUIPLOC] = GetItemInfo(itemId);

		return itemInfo[info];
	end
end

function c:GetItemName(itemId)
	local itemName = GetItemInfos(itemId, ITEMINFO_NAME);
	if itemName then
		return itemName;
	end
	return c:GetText("Empty");
end

function c:GetItemLink(itemId)
	local itemLink = GetItemInfos(itemId, ITEMINFO_LINK);
	if itemLink then
		return itemLink;
	end
	return c:GetText("Empty");
end

function c:GetItemEquipLoc(itemId)
	return GetItemInfos(itemId, ITEMINFO_EQUIPLOC);
end

local RARITY_POOR = 0;
local RARITY_COMMON = 1;
local RARITY_UNCOMMON = 2;
local RARITY_RARE = 3;
local RARITY_EPIC = 4;
local RARITY_LEGENDARY = 5;
local RARITY_ARTIFACT = 6;
local RARITY_HEIRLOOM = 7;
local RARITY_WOWTOKEN = 8;
function c:GetItemRarity(itemId)
	local itemRarity = GetItemInfos(itemId, ITEMINFO_RARITY);
	if itemRarity then
		return itemRarity;
	end
	return RARITY_COMMON;
end

function c:GetSpellName(spellId)
	local spellName = GetSpellInfo(spellId);
	if spellName then
		return spellName;
	end
	return c:GetText("Empty");
end

function c:GetSpellRank(spellId)
	return GetSpellSubtext(spellId);
end

local MACROINFO_NAME = "MACROINFO_NAME";
local MACROINFO_TEXT = "MACROINFO_TEXT";
local function GetMacroInfos(macroSlotId, info)
	local macroInfo = {};
	macroInfo[MACROINFO_NAME], macroInfo["icon"], macroInfo[MACROINFO_TEXT], macroInfo["isLocal"] = GetMacroInfo(macroSlotId);
	if MACROINFO_NAME then
		return c:GetText("Macro \"%s\"", macroInfo[MACROINFO_NAME]);
	else
		return macroInfo[info];
	end
	return c:GetText("Empty");
end

function c:GetMacroName(macroSlotId)
	return GetMacroInfos(macroSlotId, MACROINFO_NAME);
end

function c:GetMacroText(macroSlotId)
	return GetMacroInfos(macroSlotId, MACROINFO_TEXT);
end

function c:GetRarityColor(rarity)
	return ITEM_QUALITY_COLORS[rarity];
end

c.ITEM_ID = "ITEM_ID";
c.ITEM_SLOT_ID = "ITEM_SLOT_ID";
c.ITEM_BAG_ID = "BAG_ID";
function c:GetMatchingItems(slotId)
	local equipLocs, result = {}, {};
	local equipLoc = c:GetItemEquipLoc(c:GetItemId(slotId));

	if slotId == INVSLOT_MAINHAND then
		table.insert(equipLocs, "INVTYPE_2HWEAPON");
		table.insert(equipLocs, "INVTYPE_WEAPON");
	elseif slotId == INVSLOT_OFFHAND then
		table.insert(equipLocs, "INVTYPE_WEAPON");
		table.insert(equipLocs, "INVTYPE_SHIELD");
	end

	if not tContains(equipLocs, equipLoc) then
		table.insert(equipLocs, equipLoc);
	end

	for bagId = 0, NUM_BAG_SLOTS do
		for slotId = 1, GetContainerNumSlots(bagId) do
			local itemId = GetContainerItemID(bagId, slotId);
			if itemId and tContains(equipLocs, c:GetItemEquipLoc(itemId)) then
				local newEntry = {};
				newEntry[c.ITEM_BAG_ID] = bagId;
				newEntry[c.ITEM_SLOT_ID] = slotId;
				newEntry[c.ITEM_ID] = itemId;
				table.insert(result, newEntry);
			end
		end
	end
	return result, table.getn(result);
end

function c:Println(str)
	if str then
		print("[GearQuipper] " .. str);
	end
end

function c:StartsWith(str, substr, caseInsensitive)
	if caseInsensitive then
		str = string.upper(str);
		substr = string.upper(substr);
	end

   return string.sub(str, 1, string.len(substr)) == substr;
end

function c:StringContains(str, match, caseInsensitive)
	if caseInsensitive then
		str = string.upper(str);
		match = string.upper(match);
	end

	return string.find(str, "%" .. match);
end

function c:Trim(str)
	if str then
		return str:gsub("^%s*(.-)%s*$", "%1");
	end
	return "";
end

function c:BoolToText(value, trueText, falseText)
	if value then
		return trueText;
	end
	return falseText;
end

function c:GetTableSize(t)
    local count = 0
    for _, __ in pairs(t) do
        count = count + 1
    end
    return count
end

function c:TableContains(t, value, caseInsensitive)
	if not caseInsensitive then
		return tContains(t, value);
	else
		value = string.upper(value);
		for k, v in ipairs(t) do
			if string.upper(v) == value then
				return v;
			end
		end
	end
end

function c:Deepcopy(orig, copies)
    copies = copies or {}
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        if copies[orig] then
            copy = copies[orig]
        else
            copy = {}
            copies[orig] = copy
            setmetatable(copy, c:Deepcopy(getmetatable(orig), copies))
            for orig_key, orig_value in next, orig, nil do
                copy[c:Deepcopy(orig_key, copies)] = c:Deepcopy(orig_value, copies)
            end
        end
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function c:RemoveValueFromArray(array, value)
	for k, v in ipairs(array) do
		if v == value then
			return table.remove(array, k);
		end
	end
end

local actionBars = {"Action", "", "MultiBarRight", "MultiBarLeft", "MultiBarBottomRight", "MultiBarBottomLeft"};
function c:GetActionButton(slotId)
	local actionBar, buttonNo = math.floor((slotId - 1) / 12) + 1, ((slotId - 1) % 12) + 1;

	if _G[actionBars[actionBar] .. "Button" .. buttonNo] and _G[actionBars[actionBar] .. "Button" .. buttonNo]:IsVisible() then
		return _G[actionBars[actionBar] .. "Button" .. buttonNo];
	end
end

function c:GetBagSpace()
	local bagSpaceCache, sum = {}, 0;
	for i = 0, NUM_BAG_SLOTS do
		local noFreeSlots = GetContainerNumFreeSlots(i);
		bagSpaceCache[i] = noFreeSlots;
		sum = sum + noFreeSlots;
	end
	return bagSpaceCache, sum;
end

function c:CheckNeccessaryBagSpace(setName)
	local desiredSet, count, maxCount = c:LoadSet(setName), 0, 0;
	for _, slotId in ipairs(c:GetSlotSwitchOrder()) do
		local equippedItemId, desiredItemId = c:GetItemId(slotId), desiredSet[slotId];
		if (not GQ_OPTIONS[c.OPT_PARTIALSETS] or c:LoadSlotState(slotId, setName)) and equippedItemId ~= c.VALUE_NONE and equippedItemId ~= 0 and desiredItemId == c.VALUE_NONE then
			count = count + 1;
		--elseif c:GetItemId(slotId) == c.VALUE_NONE and set[slotId] and set[slotId] ~= c.VALUE_NONE then
		--	count = count - 1;
		end

		if count > maxCount then
			maxCount = count;
		end
	end

	local bagSpaceCache, freeSpace = c:GetBagSpace();
	return freeSpace, maxCount, bagSpaceCache;
end

function c:FindItemInBags(neededId)
	for bagId = 0, NUM_BAG_SLOTS do
		for slotId = 1, GetContainerNumSlots(bagId) do
			local itemId = GetContainerItemID(bagId, slotId);
			if itemId == neededId then
				return bagId, slotId;
			end
		end
	end
end

function c:IsSetComplete(setName)
	setName = setName or c:LoadCurrentSetName();
	local set = c:LoadSet(setName);
	if set then
		for slotId, itemId in pairs(set) do
			if not c:IsSetItem(slotId, itemId) then
				return false;
			end
		end
		return true;
	end
end

function c:IsSetItem(slotId, itemId)
	return c:GetItemId(slotId) == itemId;
end

function c:SaveConditionsMet()
	return GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_ALWAYS or
			(GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_CHARMENU and PaperDollFrame:IsVisible()) or
			(GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_GQMENU and GqUiFrame:IsVisible());
end

function c:IsPartialSet(setName)
	setName = setName or c:LoadCurrentSetName();
	for slotId, itemId in pairs(c:GetSlotInfo()) do
		if not c:LoadSlotState(slotId, setName) then
			return true;
		end
	end
	return false;
end

function c:GetDisplaySetName(setName)
	setName = setName or c:LoadCurrentSetName();

	local result = setName;
	if c:IsPartialSet(setName) then
		result = string.format("%s (%s)", setName, c:GetText("partial"));
	end
	return result;
end