-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

local DIALOG_ADDSET = "DIALOG_ADDSET";
local DIALOG_DELETESET = "DIALOG_DELETESET";
local DIALOG_RESET = "DIALOG_RESET";

local paperDollButton, paperDollLabel;
local slotStateBoxes;
local gqToolTip;

local secureActionButtons = {};

----- general ui stuff -----

function c:InitUI()
	paperDollButton = paperDollButton or c:CreatePaperDollButton();
	paperDollLabel = paperDollLabel or c:CreatePaperDollLabel();

	PaperDollFrame:HookScript("OnShow", function()
		if GQ_OPTIONS[c.OPT_MENUOPEN] then
			c:ToggleUI(true);
		end
		 c:SetSlotInfo();
		 c:SetPaperDollLabelText();
	end);
	PaperDollFrame:HookScript("OnHide", function()
		GQ_OPTIONS[c.OPT_MENUOPEN] = GqUiFrame.isOpen;
		c:CloseQuickBars();
	end);

	InterfaceOptionsFrameOkay:HookScript("OnClick", function(self)
		c:SaveCloakAndHelmet();
	end);

	c:CreateDialogs();
	c:InitUiFrame();

	c:CreateSecureActionButtons();
end

function c:CreatePaperDollButton()
	local button = CreateFrame("Button", "GQ_PaperDollButton", _G["PaperDollItemsFrame"]);
	button:SetPoint("TOPRIGHT", _G["PaperDollItemsFrame"], "TOPRIGHT", -40, -45);
	button:SetWidth(50);
	button:SetHeight(20);
	button:SetText("GQ");
	button:SetNormalFontObject("GameFontNormal");
	button:RegisterForClicks("LeftButtonUp");
	button:SetScript("OnClick", function() c:ToggleUI(); end);

	local ntex = button:CreateTexture();
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up");
	ntex:SetTexCoord(0, 0.625, 0, 0.6875);
	ntex:SetAllPoints();
	button:SetNormalTexture(ntex);

	local htex = button:CreateTexture();
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight");
	htex:SetTexCoord(0, 0.625, 0, 0.6875);
	htex:SetAllPoints();
	button:SetHighlightTexture(htex);

	local ptex = button:CreateTexture();
	ptex:SetTexture("Interface/Buttons/UI-Panel-Button-Down");
	ptex:SetTexCoord(0, 0.625, 0, 0.6875);
	ptex:SetAllPoints();
	button:SetPushedTexture(ptex);

	local dtex = button:CreateTexture();
	dtex:SetTexture("Interface/Buttons/UI-Panel-Button-Disabled");
	dtex:SetTexCoord(0, 0.625, 0, 0.6875);
	dtex:SetAllPoints();
	button:SetDisabledTexture(dtex);

	return button;
end

function c:DisablePaperDollButton()
	paperDollButton:Disable();
	c:EnablePaperDollButton();
end

function c:EnablePaperDollButton()
	if InCombatLockdown() then
		C_Timer.After(1, function() c:EnablePaperDollButton(); end); -- try again
	else
		paperDollButton:Enable();
	end
end

function c:CreatePaperDollLabel()
	local label = CreateFrame("Frame", "GQ_PaperDollLabel", _G["PaperDollItemsFrame"]);
	label:SetPoint("CENTER", _G["PaperDollItemsFrame"], "CENTER", 0, -5);
	label:SetWidth(200);
	label:SetHeight(20);
	label:EnableMouse();
	label:SetScript("OnEnter", function(self) self:SetAlpha(0); end);
	label:SetScript("OnLeave", function(self) self:SetAlpha(1); end);

	label.text = label:CreateFontString(nil, "ARTWORK");
	label.text:SetFont("Fonts\\ARIALN.ttf", 13, "OUTLINE");
	label.text:SetPoint("CENTER", 0, 0);

	return label;
end

function c:SetPaperDollLabelText(setName)
	if GQ_OPTIONS[c.OPT_SHOWCURRENTSET] then
		setName = setName or c:LoadCurrentSetName();
		if not setName then
			paperDollLabel.text:SetFormattedText("%s%s: %s", "|cffff0000", c:GetText("Current set"), c:GetText("None"));
		elseif c:IsSetComplete() then
			paperDollLabel.text:SetFormattedText("%s%s: \"%s\"", "|cffffcc00", c:GetText("Current set"), setName);
		else
			paperDollLabel.text:SetFormattedText("%s%s: \"%s\"", "|cffff0000", c:GetText("Current set"), setName);
		end
		paperDollLabel:Show();
	else
		paperDollLabel:Hide();
	end
end

function c:CreateDialogs()
	StaticPopupDialogs[DIALOG_ADDSET] = StaticPopupDialogs[DIALOG_ADDSET] or {
		text = c:GetText("Enter a name for this set:"),
		button1 = c:GetText("Apply"),
		button2 = c:GetText("Cancel"),
		hasEditBox = true,
		maxLetters = 30,
		EditBoxOnEnterPressed = function (self)
			local parent = self:GetParent();
			local setName = c:Trim(self:GetText());
			if c:IsValidSetName(setName) then
				c:SaveSet(setName, false);
			else
				c:Println(c:GetText("Invalid set name: \"%s\". Set names must [1] not be empty, [2] be unique (case insensitive) and [3] not contain \"$\" (reserved).", setName));
			end
			parent:Hide();
		end,
		EditBoxOnEscapePressed = function (self)
			self:GetParent():Hide();
		end,
		OnAccept = function(self, data, data2)
			local setName = c:Trim(self.editBox:GetText());
			if c:IsValidSetName(setName) then
				c:SaveSet(setName, false);
			else
				c:Println(c:GetText("Invalid set name: \"%s\". Set names must [1] not be empty, [2] be unique (case insensitive) and [3] not contain \"$\" (reserved).", setName));
			end
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
	};

	StaticPopupDialogs[DIALOG_DELETESET] = StaticPopupDialogs[DIALOG_DELETESET] or {
		text = c:GetText("Do you really want to delete \"%s\"?", "%s"),
		button1 = c:GetText("Apply"),
		button2 = c:GetText("Cancel"),
		EditBoxOnEscapePressed = function (self)
			self:GetParent():Hide();
		end,
		OnAccept = function(self, data)
			if data and tContains(c:LoadSetNames(), data) then
				c:DeleteSet(data);
			end
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
	};

	StaticPopupDialogs[DIALOG_RESET] = StaticPopupDialogs[DIALOG_RESET] or {
		text = c:GetText("This will reset GearQuipper to defaults. All your sets will be lost.\n\nAre you sure?"),
		button1 = c:GetText("Reset"),
		button2 = c:GetText("Cancel"),
		OnEscapePressed = function ()
			self:Hide();
		end,
		OnAccept = function()
			GQ_OPTIONS = {};
			GQ_DATA = {};
			c:ToggleUI(false);
			c:Init();
			c:SaveDefaultSet();
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
	};
end

function c:ShowAddSetDialog()
	StaticPopup_Show(DIALOG_ADDSET);
end

function c:ShowDeleteSetDialog(setName)
	if setName then
		local dialog = StaticPopup_Show(DIALOG_DELETESET, setName);
		if dialog then
			dialog.data = setName;
		end
	end
end

function c:ShowResetDialog()
	StaticPopup_Show(DIALOG_RESET);
end

function c:SetSlotInfo(slotId, color)
	if not slotId and not color then
		local set = c:LoadSet();
		if set then
			for slotId, slotName in pairs(c:GetSlotInfo()) do
				if c:IsSetItem(slotId, set[slotId]) then
					c:SetSlotInfo(slotId, "green");
				else
					c:SetSlotInfo(slotId, "red");
				end
			end
		end
		return;
	end

	local slot = _G["Character" .. c:GetSlotInfo()[slotId]];
	if not slot.gqTex then
		slot.slotId = slotId;
		slot.gqTex = slot:CreateTexture(nil, "OVERLAY");
		slot.gqTex:SetAllPoints();
		slot:HookScript("OnEnter", function(self)
			c:InjectTooltip(self);
		end);

		slot:HookScript("OnLeave", function()
			if gqToolTip then
				gqToolTip:Hide();
			end
		end);
	end

	if color == "red" then
		slot.gqTex:SetColorTexture(1, 0, 0, 0.2);
		slot.gqInfo = c:GetText("|cffff0000Expected Item:|r %s", c:GetItemLink(c:LoadSet()[slotId]));
	else
		slot.gqTex:SetColorTexture(0, 0, 0, 0);
		slot.gqInfo = nil;
	end
end

function c:InjectTooltip(slot)
	-- experimental
	if not gqToolTip then
		gqToolTip = CreateFrame("GameTooltip", "GQ_SlotTooltip", UIParent, "GameTooltipTemplate");
	end

	gqToolTip:SetOwner(slot, "ANCHOR_RIGHT");
	if slot.gqInfo then
		gqToolTip:AddLine(slot.gqInfo);
		gqToolTip:Show();

		GameTooltip:SetOwner(gqToolTip, "ANCHOR_BOTTOM");
		GameTooltip:SetHyperlink(c:GetItemLink(c:GetItemId(slot.slotId)));
	end
end

function c:GetSlotStateBoxes()
	if not slotStateBoxes then
		slotStateBoxes = {};
		for slotId, slotName in pairs(c:GetSlotInfo()) do
			slotStateBoxes[slotId] = c:CreateSlotStateBox(slotName, _G["Character" .. slotName]);
		end
	end
	return slotStateBoxes;
end

function c:ShowSlotStateBoxes()
	if GQ_OPTIONS[c.OPT_PARTIALSETS] and GqUiFrame:IsVisible() then
		for slotId, checkBox in pairs(c:GetSlotStateBoxes()) do
			c:SetSlotState(slotId, c:LoadSlotState(slotId));
			checkBox:Show();
		end
	end
end

function c:HideSlotStateBoxes()
	for slotId, checkBox in pairs(c:GetSlotStateBoxes()) do
		checkBox:Hide();
	end
end

function c:CreateSlotStateBox(slotName, parent)
	local checkbutton = CreateFrame("CheckButton", slotName .. "StateBox", parent, "UICheckButtonTemplate");
	checkbutton:SetWidth(24);
	checkbutton:SetHeight(24);
	checkbutton:SetPoint("BOTTOMLEFT", -4, -4);
	checkbutton:RegisterForClicks("LeftButtonUp");
	checkbutton:SetScript("OnClick",
		function(self)
			local slotId = c:GetSlotId(slotName);
			c:SaveSlotState(slotId, self:GetChecked());
			if GqUiFrame:IsVisible() then
				c:RefreshSetList();
			end

			if not self:GetChecked() then
				c:SaveSlot(slotId, c:LoadCurrentSetName(), true, c.VALUE_NONE);
			elseif not c:IsSetItem(slotId, c:LoadSet()[slotId]) then
				c:SaveSlot(slotId);
			end

			c:SetSlotInfo(slotId, "green");
		end);
	return checkbutton;
end

function c:SetSlotState(slotId, value)
	c:GetSlotStateBoxes()[slotId]:SetChecked(value);
end

function c:GetSlotState(slotId)
	return c:GetSlotStateBoxes()[slotId]:GetChecked();
end

----- experimental -----

function c:CreateSecureActionButtons()
	for _, setName in ipairs(c:LoadSetNames()) do
		local set = c:LoadSet(setName);
		local macrotext;

		if set[INVSLOT_MAINHAND] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_MAINHAND);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_MAINHAND, c:GetItemName(set[INVSLOT_MAINHAND]));
		end

		if set[INVSLOT_OFFHAND] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_OFFHAND);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_OFFHAND, c:GetItemName(set[INVSLOT_OFFHAND]));
		end

		if set[INVSLOT_RANGED] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_RANGED);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_RANGED, c:GetItemName(set[INVSLOT_RANGED]));
		end

		c:SetSecureAction(setName, macrotext);
	end
end

function c:CreateSecureActionButton(setName)
	if setName then
		if not secureActionButtons[setName] then
			local button = CreateFrame("Button", "GQ_SecureActionButton_" .. setName, UIParent, "SecureActionButtonTemplate");
			button:RegisterForClicks("AnyUp");
			button:SetAttribute("type", "macro");
			secureActionButtons[setName] = button;
		end
		return secureActionButtons[setName];
	end
end

function c:SetSecureAction(setName, macrotext)
	if setName and macrotext then
		local button = c:CreateSecureActionButton(setName);
		button:SetAttribute("macrotext", macrotext);
	end
end

function c:CallSecureAction(setName)
	if secureActionButtons[setName] then
		-- experimental
		--C_Timer.After(0.1, function()
		--	secureActionButtons[setName]:Click();
		--end);
	end
end