-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

local optionElementIndex = 1;
local optionElements, setCheckBoxes, eventEntries = {}, {}, {};
local checkboxYOffset, dropdownYOffset = 22, 27;

local labelSaveChanges;
local cbSave_Always, cbSave_CharMenu, cbSave_GqMenu;

local bindingEntries, bindingEntriesOrder;
local TEXT_DROPDOWN, TEXT_BUTTON, TEXT_CHAT = "TEXT_DROPDOWN", "TEXT_BUTTON", "TEXT_CHAT";

----- ui frame general stuff -----

function c:InitUiFrame()
	GqUiFrameFontString:SetText("GearQuipper Version " .. c:GetVersion());
	GqPreferencesFrameFontString:SetText(c:GetText("Preferences"));
	GqUiFrame_BtnSaveSet:SetText(c:GetText("Save set"));
	GqUiFrame_BtnCreateNewSet:SetText(c:GetText("Add set..."));
	GqUiFrame_BtnDeleteSet:SetText(c:GetText("Remove set"));
	GqUiFrame_BtnPreferences:SetText(c:GetText("Preferences"));

	optionElementIndex = 1;
	c:CreateOptionElements();

	GqUiFrame.CloseButton:SetScript("OnClick", function(self) c:ToggleUI(false); end);
end

function c:ToggleUI(value)
	if c.initFinished then
		if value or not GqUiFrame:IsVisible() then
			GqUiFrame:Show();
			GqUiFrame.isOpen = true;
		else
			GqUiFrame:Hide();
			GqUiFrame.isOpen = false;
		end
	end
end

function c:TogglePreferences(value)
	if value or not GqPreferencesFrame:IsVisible() then
		GqPreferencesFrame:Show();
	else
		GqPreferencesFrame:Hide();
	end
end

function c:LockUI(setName)
	if PaperDollFrame:IsVisible() then
		c:HideSlotStateBoxes();
	end

	if GqUiFrame:IsVisible() then
		if setCheckBoxes then
			for index, checkBox in ipairs(setCheckBoxes) do
				checkBox:SetEnabled(false);
				checkBox:SetChecked(false);
				if checkBox.setName ~= setName then
					checkBox:SetAlpha(0.5);
				end
			end
		end

		GqUiFrame_BtnSaveSet:SetEnabled(false);
		GqUiFrame_BtnCreateNewSet:SetEnabled(false);
		GqUiFrame_BtnDeleteSet:SetEnabled(false);
		GqUiFrame_BtnPreferences:SetEnabled(false);

		if optionElements then
			for index, optionElement in pairs(optionElements) do
				optionElement:SetEnabled(false);
			end
		end

		if eventEntries then
			for index, entry in ipairs(eventEntries) do
				if entry.dropdown then
					UIDropDownMenu_DisableDropDown(entry.dropdown);
				end
			end
		end
	end
end

function c:UnlockUI()
	c:ShowSlotStateBoxes();

	GqUiFrame_BtnSaveSet:SetEnabled(true);
	GqUiFrame_BtnCreateNewSet:SetEnabled(true);
	GqUiFrame_BtnDeleteSet:SetEnabled(true);
	GqUiFrame_BtnPreferences:SetEnabled(true);

	if optionElements then
		for index, optionElement in pairs(optionElements) do
			optionElement:SetEnabled(true);
		end
	end

	if eventEntries then
		for index, entry in ipairs(eventEntries) do
			if entry.dropdown then
				UIDropDownMenu_EnableDropDown(entry.dropdown);
			end
		end
	end
end

function c:CreateOptionElements()
	-- save changes options
	if not labelSaveChanges then
		labelSaveChanges = CreateFrame("Frame", "GqUiOptionElement_LabelSaveChanges", GqPreferencesFrame);
		labelSaveChanges:SetPoint("TOPLEFT", 35, -(optionElementIndex - 1) * checkboxYOffset - 30);
		labelSaveChanges:SetWidth(100);
		labelSaveChanges:SetHeight(23);
		labelSaveChanges.fontString = labelSaveChanges:CreateFontString("GqUiOptionElement_LabelSaveChanges_FontString", "ARTWORK", "GameFontNormal");
		labelSaveChanges.fontString:SetPoint("LEFT", 3, 0);
		labelSaveChanges.fontString:SetText(c:GetText("Save changes when:"));
	end

	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Always"), BtnSave_Always_OnClick, 175);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_ALWAYS);
	cbSave_Always = optionElements[optionElementIndex];
	optionElementIndex = optionElementIndex + 1;

	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Character menu open"), BtnSave_CharMenu_OnClick, 175);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_CHARMENU);
	cbSave_CharMenu = optionElements[optionElementIndex];
	optionElementIndex = optionElementIndex + 1;

	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("GearQuipper menu open"), BtnSave_GqMenu_OnClick, 175);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_SAVECHANGES] == c.OPTVALUE_SAVECHANGES_GQMENU);
	cbSave_GqMenu = optionElements[optionElementIndex];
	optionElementIndex = optionElementIndex + 1;

	-- switch action bars option
	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Save/Load action slots with your sets"), BtnSwitchActionBars_OnClick);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_ACTIONBARS]);
	optionElementIndex = optionElementIndex + 1;

	--  partial sets option
	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Enable partial sets"), BtnPartialSets_OnClick);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_PARTIALSETS]);
	optionElementIndex = optionElementIndex + 1;

	-- highlight changes option
	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Highlight switched slots"), BtnHighlightSlots_OnClick);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES]);
	optionElementIndex = optionElementIndex + 1;

	-- notify changes option
	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Confirm changes in chat"), BtnConfirmChanges_OnClick);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_NOTIFYCHANGES]);
	optionElementIndex = optionElementIndex + 1;

	-- notify changes option
	optionElements[optionElementIndex] = optionElements[optionElementIndex] or c:CreateOptionCheckButton(c:GetText("Show \"current set\" label"), BtnShowCurrentSet_OnClick);
	optionElements[optionElementIndex]:SetChecked(GQ_OPTIONS[c.OPT_NOTIFYCHANGES]);
	optionElementIndex = optionElementIndex + 1;

	-- reset gearquipper option
	if not optionElements[0] then
		optionElements[0] = CreateFrame("Button", "GqUiOptionButton_" .. optionElementIndex, GqPreferencesFrame, "GqButtonTemplate");
		optionElements[0]:SetWidth(180);
		optionElements[0]:SetPoint("BOTTOMRIGHT", -12, 10);
		optionElements[0]:SetScript("OnClick", function() c:ShowResetDialog(); end);
		optionElements[0].fontString = optionElements[0]:CreateFontString("GqUiSetButton_" .. optionElementIndex .. "_FontString", "ARTWORK", "GameFontNormal")
		optionElements[0].fontString:SetPoint("CENTER");
		optionElements[0].fontString:SetText(c:GetText("Reset GearQuipper"));
	end
end

function c:CreateOptionCheckButton(text, func, xOffset)
	xOffset = xOffset or 0;

	local element = CreateFrame("CheckButton", "GqUiOptionElement_" .. optionElementIndex, GqPreferencesFrame, "ChatConfigCheckButtonTemplate");
	element:SetPoint("TOPLEFT", 35 + xOffset, -(optionElementIndex - 1) * checkboxYOffset - 30);
	element:SetScript("OnClick", function(self) func(self); end);

	element.fontString = element:CreateFontString("GqUiOptionElement_" .. optionElementIndex .. "_FontString", "ARTWORK", "GameFontNormal");
	element.fontString:SetPoint("LEFT", 30, 0);
	element.fontString:SetText(text);

	return element;
end

function c:RefreshSetList(scrollIntoView)
	-- hide boxes
	for index, checkBox in ipairs(setCheckBoxes) do
		checkBox:Hide();
	end

	-- create boxes if neccessary and unlock
	local setNames, currentSet, currentSetPosition = c:LoadSetNames(), c:LoadCurrentSetName();
	for index, setName in ipairs(setNames) do
		setCheckBoxes[index] = setCheckBoxes[index] or c:CreateSetEntry(index);
		setCheckBoxes[index].setName = setName;
		setCheckBoxes[index].fontString:SetText(c:GetDisplaySetName(setName));
		setCheckBoxes[index]:SetChecked(setName == currentSet);
		setCheckBoxes[index]:SetAlpha(1);
		setCheckBoxes[index]:SetEnabled(true);
		setCheckBoxes[index]:Show();

		if scrollIntoView and setName == currentSet then
			currentSetPosition = (index - 1) * checkboxYOffset;
		end
	end

	local noSets = table.getn(setNames);
	if noSets > 5 then
		local maxValue = (noSets * checkboxYOffset) - GqUiScrollFrame:GetHeight() + 1;
		GqUiScrollFrameVSlider:SetMinMaxValues(0, maxValue);
		GqUiScrollFrameVSlider:Show();

		if currentSetPosition then
			GqUiScrollFrameVSlider:SetValue(currentSetPosition);
		end
	elseif GqUiScrollFrameVSlider:IsVisible() then
		GqUiScrollFrameVSlider:Hide();
	end
end

function c:CreateSetEntry(index)
	local button = CreateFrame("CheckButton", "GqUiSetButton_" .. index, GqUiScrollFrameContent, "ChatConfigCheckButtonTemplate");
	button:SetPoint("TOPLEFT", 0, -(index - 1) * checkboxYOffset);
	button:SetScript("OnClick", function(self) SetCheckBox_OnClick(self); end);

	button.fontString = button:CreateFontString("GqUiSetButton_" .. index .. "_FontString", "ARTWORK", "GameFontNormal")
	button.fontString:SetPoint("LEFT", 30,0);

	return button;
end

function c:GetBindingEntries(refresh)
	if refresh then
		bindingEntries, bindingEntriesOrder = nil, nil;
	end

	if not bindingEntries then
		bindingEntries = {
			[c.KEYWORD_NONE] = {
				TEXT_DROPDOWN = ("< " .. c:GetText("None") .. " >"),
				TEXT_CHAT = c:GetText("None")
			},
			[c.KEYWORD_PREVIOUS] = {
				TEXT_DROPDOWN = ("< " .. c:GetText("Previous set") .. " >"),
				TEXT_BUTTON = c:GetText("Previous set"),
				TEXT_CHAT = c:GetText("Previous set")
			},
		};

		for index, setName in ipairs(c:LoadSetNames()) do
			bindingEntries[setName] = setName;
		end
	end

	if not bindingEntriesOrder then
		bindingEntriesOrder = {
			c.KEYWORD_NONE,
			c.KEYWORD_PREVIOUS
		};
		for index, setName in ipairs(c:LoadSetNames()) do
			table.insert(bindingEntriesOrder, setName);
		end
	end

	return bindingEntries, bindingEntriesOrder;
end

function c:GetBindingText(bindingName, textType)
	if bindingName then
		local bindings = c:GetBindingEntries();
		if bindings[bindingName] then
			return bindings[bindingName][textType] or bindings[bindingName];
		end
	end
end

function c:RefreshEventEntries()
	for index, eventName in ipairs(c:GetEventNamesSorted()) do
		eventEntries[index] = eventEntries[index] or c:CreateEventEntry(index, eventName);
		UIDropDownMenu_SetText(eventEntries[index].dropdown, c:GetBindingText(c:LoadEventBinding(eventName), TEXT_BUTTON));
	end

	local noEvents = table.getn(c:GetEventNamesSorted());
	noEvents = noEvents + 1;
	eventEntries[noEvents] = eventEntries[noEvents] or c:CreateEventEntry(noEvents, "z1"); -- todo: remove
	noEvents = noEvents + 1;
	eventEntries[noEvents] = eventEntries[noEvents] or c:CreateEventEntry(noEvents, "z2"); -- todo: remove
	
	if noEvents > 5 then
		local maxValue = (noEvents * dropdownYOffset) - GqUiEventBindingsScrollFrame:GetHeight() + 5;
		GqUiEventBindingsScrollFrameVSlider:SetMinMaxValues(0, maxValue);
		GqUiEventBindingsScrollFrameVSlider:Show();
	elseif GqUiEventBindingsScrollFrameVSlider:IsVisible() then
		GqUiEventBindingsScrollFrameVSlider:Hide();
	end
end

function c:CreateEventEntry(index, eventName)
	local frame = CreateFrame("Frame", "GqUiBindingEntry_" .. index, GqUiEventBindingsScrollFrameContent);
	frame:SetPoint("TOPLEFT", 5, -(index - 1) * dropdownYOffset - 5);
	frame:SetHeight(27);
	frame:SetWidth(frame:GetParent():GetWidth() + 6);

	frame.label = CreateFrame("Frame", "GqUiBindingEntry_" .. index .. "_Label", frame);
	frame.label:SetPoint("TOPLEFT");
	frame.label:SetHeight(22);
	frame.label:SetWidth(frame:GetParent():GetWidth());
	frame.label.fontString = frame.label:CreateFontString("GqUiBindingEntry_" .. index .. "_Label_FontString", "ARTWORK", "GameFontNormalSmall")
	frame.label.fontString:SetPoint("LEFT");
	frame.label.fontString:SetText(c:GetEvents()[eventName]);

	if index <= 4 then -- todo: remove later
		frame.dropdown = CreateFrame("Frame", "GqUiBindingEntry_" .. index .. "_Dropdown", frame, "UIDropDownMenuTemplate");
		frame.dropdown:SetPoint("RIGHT", -10, 0);
		UIDropDownMenu_SetWidth(frame.dropdown, 90);

		UIDropDownMenu_Initialize(frame.dropdown, function(self, level, menuList)
			local info, setNames = UIDropDownMenu_CreateInfo(), c:LoadSetNames();
			info.hasArrow = false;
			info.notCheckable  = true;
			info.func = function(self)
				if c:BindEvent(eventName, self.value) then
					UIDropDownMenu_SetText(frame.dropdown, c:GetBindingText(self.value, TEXT_BUTTON));
				end
				CloseDropDownMenus();
			end;

			if (level or 1) == 1 then
				local bindings, order = c:GetBindingEntries();
				for index, entry in ipairs(order) do
					info.text = c:GetBindingText(entry, TEXT_DROPDOWN);
					info.value = entry;
					UIDropDownMenu_AddButton(info);
				end
			end
		end);
	end

	return frame;
end

function c:BindEvent(eventName, setName)
	if eventName and setName ~= c:LoadEventBinding(eventName) then
		c:SaveEventBinding(eventName, setName);
		return true;
	end
end

----- ui frame interaction -----

function GqUiFrame_OnShow(self)
	if c:IsSwitching() then
		c:LockUI();
	else
		c:UnlockUI();
	end

	c:ShowSlotStateBoxes();
end

function GqUiFrame_OnHide(self)
	c:HideSlotStateBoxes();
end

function GqUiScrollFrame_OnShow(self)
	c:RefreshSetList();
end

function GqUiScrollFrame_OnMouseWheel(self, delta)
	if GqUiScrollFrameVSlider:IsVisible() then
		local current, min, max = GqUiScrollFrameVSlider:GetValue(), GqUiScrollFrameVSlider:GetMinMaxValues();
		if delta < 0 and current < max then
			GqUiScrollFrameVSlider:SetValue(current + max / 10);
		elseif (delta > 0) and (current > 1) then
			GqUiScrollFrameVSlider:SetValue(current - max / 10);
		end
	end
end

function GqUiEventBindingsScrollFrame_OnShow(self)
	c:RefreshEventEntries();
end

function GqUiEventBindingsScrollFrame_OnMouseWheel(self, delta)
	if GqUiEventBindingsScrollFrameVSlider:IsVisible() then
		local current, min, max = GqUiEventBindingsScrollFrameVSlider:GetValue(), GqUiEventBindingsScrollFrameVSlider:GetMinMaxValues();
		if delta < 0 and current < max then
			GqUiEventBindingsScrollFrameVSlider:SetValue(current + max / 10);
		elseif (delta > 0) and (current > 1) then
			GqUiEventBindingsScrollFrameVSlider:SetValue(current - max / 10);
		end
	end
end

function SetCheckBox_OnClick(self)
	c:SwitchToSet(self.setName);
end

function BtnSaveSet_OnClick()
	local setName = c:LoadCurrentSetName();
	if setName then
		c:SaveSet(setName, true);
		c:SetPaperDollLabelText();
		c:SetSlotInfo();
	end
end

function BtnCreateNewSet_OnClick()
	c:ShowAddSetDialog();
end

function BtnDeleteSet_OnClick()
	local setName = c:LoadCurrentSetName();
	if setName then
		c:ShowDeleteSetDialog(setName);
	end
end

function BtnPreferences_OnClick()
	c:TogglePreferences();
end

function BtnSwitchActionBars_OnClick(self)
	GQ_OPTIONS[c.OPT_ACTIONBARS] = self:GetChecked();
end

function BtnConfirmChanges_OnClick(self)
	GQ_OPTIONS[c.OPT_NOTIFYCHANGES] = self:GetChecked();
end

function BtnHighlightSlots_OnClick(self)
	GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] = self:GetChecked();
end

function BtnPartialSets_OnClick(self)
	GQ_OPTIONS[c.OPT_PARTIALSETS] = self:GetChecked();
	if GQ_OPTIONS[c.OPT_PARTIALSETS] then
		c:ShowSlotStateBoxes();
	else
		c:HideSlotStateBoxes();
	end
end

function BtnShowCurrentSet_OnClick(self)
	GQ_OPTIONS[c.OPT_SHOWCURRENTSET] = self:GetChecked();
	c:SetPaperDollLabelText();
end

function BtnSave_Always_OnClick(self)
	if self:GetChecked() then
		GQ_OPTIONS[c.OPT_SAVECHANGES] = c.OPTVALUE_SAVECHANGES_ALWAYS;
		cbSave_CharMenu:SetChecked(false);
		cbSave_GqMenu:SetChecked(false);
	else
		self:SetChecked(true);
	end
end

function BtnSave_CharMenu_OnClick(self)
	if self:GetChecked() then
		GQ_OPTIONS[c.OPT_SAVECHANGES] = c.OPTVALUE_SAVECHANGES_CHARMENU;
		cbSave_Always:SetChecked(false);
		cbSave_GqMenu:SetChecked(false);
	else
		self:SetChecked(true);
	end
end

function BtnSave_GqMenu_OnClick(self)
	if self:GetChecked() then
		GQ_OPTIONS[c.OPT_SAVECHANGES] = c.OPTVALUE_SAVECHANGES_GQMENU;
		cbSave_Always:SetChecked(false);
		cbSave_CharMenu:SetChecked(false);
	else
		self:SetChecked(true);
	end
end